#include "bmp.h"
#include <windows.h>
#include <cstdio>
#include <cstdlib>

bool image_realloc(imagestruct* image) {
	unsigned short y,yy;
	image->data = (pixel**) malloc(image->height*sizeof(pixel*));
	if(image->data==NULL) {
		return false;
	}
	for(y=0;y<image->height;++y) {
		image->data[y] = (pixel*) malloc(image->width*sizeof(pixel));
		if(image->data[y]==NULL) {
			for(yy=0;yy<y;++yy) {
				free(image->data[yy]);
			}
			free(image->data);
			return false;
		}
	}
	return true;
}

bool image_load_empty(imagestruct* image, unsigned short width, unsigned short height, unsigned char r, unsigned char g, unsigned char b, unsigned char a) {
	
	unsigned short x,y;
	
	image->width = width;
	image->height = height;
	if(!image_realloc(image)) {
		return false;
	}
	
	for(y=0;y<image->height;++y) {
		for(x=0;x<image->width;++x) {
			image->data[y][x].r = r;
			image->data[y][x].g = g;
			image->data[y][x].b = b;
			image->data[y][x].a = a;
		}
	}
	
	return true;
	
}

bool image_load_bmp(imagestruct* image, const char* filename) {
	
	FILE *f;
	BITMAPFILEHEADER fileheader;
	BITMAPINFOHEADER infoheader;
	unsigned long size;
	char *data;
	unsigned short rowlength;
	unsigned short x,y;
	
	f = fopen(filename,"rb");
	if(f==NULL) {
		return false;
	}
	
	if(fread(&fileheader,sizeof(fileheader),1,f)!=1) {
		fclose(f);
		return false;
	}
	if(fileheader.bfType!=0x4D42) {
		fclose(f);
		return false;
	}
	
	if(fread(&infoheader,sizeof(infoheader),1,f)!=1) {
		fclose(f);
		return false;
	}
	
	fseek(f,fileheader.bfOffBits,SEEK_SET);
	data = (char*) malloc(infoheader.biSizeImage);
	if(data==NULL) {
		fclose(f);
		return false;
	}
	if(fread(data,1,infoheader.biSizeImage,f)!=infoheader.biSizeImage) {
		free(data);
		fclose(f);
		return false;
	}
	fclose(f);
	
	if(infoheader.biBitCount==24) {
		image->width = abs(infoheader.biWidth);
		image->height = abs(infoheader.biHeight);
		if(!image_realloc(image)) {
			free(data);
			return false;
		}
		rowlength = infoheader.biWidth*3;
		if((rowlength%4)!=0) rowlength += 4-(rowlength%4);
		if(infoheader.biHeight<0) {
			for(y=0;y<image->height;++y) {
				for(x=0;x<image->width;++x) {
					image->data[y][x].r = data[y*rowlength+x*3+2];
					image->data[y][x].g = data[y*rowlength+x*3+1];
					image->data[y][x].b = data[y*rowlength+x*3+0];
					image->data[y][x].a = 255;
				}
			}
		} else {
			for(y=0;y<image->height;++y) {
				for(x=0;x<image->width;++x) {
					image->data[y][x].r = data[(infoheader.biHeight-1-y)*rowlength+x*3+2];
					image->data[y][x].g = data[(infoheader.biHeight-1-y)*rowlength+x*3+1];
					image->data[y][x].b = data[(infoheader.biHeight-1-y)*rowlength+x*3+0];
					image->data[y][x].a = 255;
				}
			}
		}
	} else if(infoheader.biBitCount==32) {
		image->width = abs(infoheader.biWidth);
		image->height = abs(infoheader.biHeight);
		if(!image_realloc(image)) {
			free(data);
			return false;
		}
		rowlength = infoheader.biWidth*4;
		if((rowlength%4)!=0) rowlength += 4-(rowlength%4);
		if(infoheader.biHeight<0) {
			for(y=0;y<image->height;++y) {
				for(x=0;x<image->width;++x) {
					image->data[y][x].r = data[y*rowlength+x*4+2];
					image->data[y][x].g = data[y*rowlength+x*4+1];
					image->data[y][x].b = data[y*rowlength+x*4+0];
					image->data[y][x].a = data[y*rowlength+x*4+3];
				}
			}
		} else {
			for(y=0;y<image->height;++y) {
				for(x=0;x<image->width;++x) {
					image->data[y][x].r = data[(infoheader.biHeight-1-y)*rowlength+x*4+2];
					image->data[y][x].g = data[(infoheader.biHeight-1-y)*rowlength+x*4+1];
					image->data[y][x].b = data[(infoheader.biHeight-1-y)*rowlength+x*4+0];
					image->data[y][x].a = data[(infoheader.biHeight-1-y)*rowlength+x*4+3];
				}
			}
		}
	} else {
		free(data);
		return false;
	}
	
	free(data);
	return true;
	
}

bool image_save_bmp(imagestruct* image, const char* filename, bool savealpha) {
	
	FILE *f;
	BITMAPFILEHEADER fileheader;
	BITMAPINFOHEADER infoheader;
	unsigned long size;
	char *data;
	unsigned short rowlength;
	unsigned short x,y;
	
	if(savealpha) rowlength = image->width*4;
	else rowlength = image->width*3;
	
	if((rowlength%4)!=0) rowlength += 4-(rowlength%4);
	
	fileheader.bfType = 0x4D42;
	fileheader.bfSize = sizeof(fileheader)+sizeof(infoheader)+rowlength*image->height;
	fileheader.bfReserved1 = 0;
	fileheader.bfReserved2 = 0;
	fileheader.bfOffBits = sizeof(fileheader)+sizeof(infoheader);
	
	infoheader.biSize = sizeof(infoheader);
	infoheader.biWidth = image->width;
	infoheader.biHeight = image->height;
	infoheader.biPlanes = 1;
	if(savealpha) infoheader.biBitCount = 32;
	else infoheader.biBitCount = 24;
	infoheader.biCompression = 0;
	infoheader.biSizeImage = rowlength*image->height;
	infoheader.biXPelsPerMeter = 2835;
	infoheader.biYPelsPerMeter = 2835;
	infoheader.biClrUsed = 0;
	infoheader.biClrImportant = 0;
	
	data = (char*) malloc(infoheader.biSizeImage);
	if(data==NULL) {
		return false;
	}
	
	if(savealpha) {
		for(y=0;y<image->height;++y) {
			for(x=0;x<image->width;++x) {
				data[(infoheader.biHeight-1-y)*rowlength+x*4+2] = image->data[y][x].r;
				data[(infoheader.biHeight-1-y)*rowlength+x*4+1] = image->data[y][x].g;
				data[(infoheader.biHeight-1-y)*rowlength+x*4+0] = image->data[y][x].b;
				data[(infoheader.biHeight-1-y)*rowlength+x*4+3] = image->data[y][x].a;
			}
		}
	} else {
		for(y=0;y<image->height;++y) {
			for(x=0;x<image->width;++x) {
				data[(infoheader.biHeight-1-y)*rowlength+x*3+2] = image->data[y][x].r;
				data[(infoheader.biHeight-1-y)*rowlength+x*3+1] = image->data[y][x].g;
				data[(infoheader.biHeight-1-y)*rowlength+x*3+0] = image->data[y][x].b;
			}
		}
	}
	
	f = fopen(filename,"wb");
	if(fwrite(&fileheader,sizeof(fileheader),1,f)!=1) {
		fclose(f);
		free(data);
		return false;
	}
	if(fwrite(&infoheader,sizeof(infoheader),1,f)!=1) {
		fclose(f);
		free(data);
		return false;
	}
	if(fwrite(data,1,infoheader.biSizeImage,f)!=infoheader.biSizeImage) {
		fclose(f);
		free(data);
		return false;
	}
	fclose(f);
	
	free(data);
	
	return true;
	
}

void image_free(imagestruct* image) {
	unsigned short y;
	for(y=0;y<image->height;++y) {
		free(image->data[y]);
	}
	free(image->data);
	image->data = NULL;
}

