#include "dll.h"
#include <cstdio>
#include <cstdlib>
#include <cmath>

// ########## BEGIN variables ##########

unsigned long id_textures = 0;
gmrender_texture *textures = NULL;

unsigned long id_models = 0;
gmrender_model *models = NULL;

float cam_longitude,cam_latitude,cam_longitude_incr,cam_scale;
unsigned short cam_orig_x,cam_orig_y,cam_width,cam_height;
unsigned char light_ambient_r,light_ambient_g,light_ambient_b;
unsigned char light_directional_r,light_directional_g,light_directional_b;
float light_directional_longitude,light_directional_latitude,light_directional_longitude_incr;
unsigned short aa_samples;
float aa_jitter;
unsigned long aa_seed;
bool aa_compalpha,aa_setedgepixels;
unsigned char aa_bg_r,aa_bg_g,aa_bg_b;

// ########## BEGIN basic functions ##########

gmrender_texture* search_texture(unsigned long id) {
	gmrender_texture *ptr;
	for(ptr=textures;ptr!=NULL;ptr=ptr->next) {
		if(ptr->id==id) return ptr;
	}
	return NULL;
}
gmrender_model* search_model(unsigned long id) {
	gmrender_model *ptr;
	for(ptr=models;ptr!=NULL;ptr=ptr->next) {
		if(ptr->id==id) return ptr;
	}
	return NULL;
}

gmrender_texture* create_texture() {
	gmrender_texture *ptr;
	if(textures==NULL) {
		textures = (gmrender_texture*) malloc(sizeof(gmrender_texture));
		if(textures==NULL) return NULL;
		textures->next = NULL;
		return textures;
	} else {
		for(ptr=textures;ptr->next!=NULL;ptr=ptr->next);
		ptr->next = (gmrender_texture*) malloc(sizeof(gmrender_texture));
		if(ptr->next==NULL) return NULL;
		ptr->next->next = NULL;
		return ptr->next;
	}
}
gmrender_model* create_model() {
	gmrender_model *ptr;
	if(models==NULL) {
		models = (gmrender_model*) malloc(sizeof(gmrender_model));
		if(models==NULL) return NULL;
		models->next = NULL;
		return models;
	} else {
		for(ptr=models;ptr->next!=NULL;ptr=ptr->next);
		ptr->next = (gmrender_model*) malloc(sizeof(gmrender_model));
		if(ptr->next==NULL) return NULL;
		ptr->next->next = NULL;
		return ptr->next;
	}
}

void destroy_texture(gmrender_texture* texture) {
	gmrender_texture *ptr;
	if(textures==texture) {
		textures = texture->next;
		free(texture);
	} else {
		for(ptr=textures;ptr->next!=texture;ptr=ptr->next);
		ptr->next = texture->next;
		free(texture);
	}
}
void destroy_model(gmrender_model* model) {
	gmrender_model *ptr;
	if(models==model) {
		models = model->next;
		free(model);
	} else {
		for(ptr=models;ptr->next!=model;ptr=ptr->next);
		ptr->next = model->next;
		free(model);
	}
}

// ########## BEGIN normal functions ##########

signed short read_signed_short(unsigned char* data, unsigned short offset) {
	return ((data[offset]<<8)|data[offset+1])-32768;
}
signed char read_signed_char(unsigned char* data, unsigned short offset) {
	return (data[offset])-128;
}
unsigned char read_unsigned_char(unsigned char* data, unsigned short offset) {
	return data[offset];
}

bool gmbin_load(gmrender_triangle** triangles, unsigned short* trianglecount, const char* filename) {
	unsigned char header[4];
	unsigned char data[51];
	void* temp;
	FILE *f;
	
	f = fopen(filename,"rb");
	if(f==NULL) {
		return false;
	}
	
	if(fread(header,1,sizeof(header),f)!=sizeof(header)) {
		fclose(f);
		return false;
	}
	if(header[0]!=251 || header[1]!=84 || header[2]!=165 || header[3]!=109) {
		fclose(f);
		return false;
	}
	
	*triangles = NULL;
	*trianglecount = 0;
	while(fread(data,1,sizeof(data),f)==sizeof(data)) {
		temp = realloc(*triangles,(*trianglecount+1)*sizeof(gmrender_triangle));
		if(temp==NULL) {
			free(*triangles);
			fclose(f);
			return false;
		}
		*triangles = (gmrender_triangle*) temp;
		(*triangles)[*trianglecount].x1 = read_signed_short(data,0);
		(*triangles)[*trianglecount].y1 = read_signed_short(data,2);
		(*triangles)[*trianglecount].z1 = read_signed_short(data,4);
		(*triangles)[*trianglecount].x2 = read_signed_short(data,6);
		(*triangles)[*trianglecount].y2 = read_signed_short(data,8);
		(*triangles)[*trianglecount].z2 = read_signed_short(data,10);
		(*triangles)[*trianglecount].x3 = read_signed_short(data,12);
		(*triangles)[*trianglecount].y3 = read_signed_short(data,14);
		(*triangles)[*trianglecount].z3 = read_signed_short(data,16);
		(*triangles)[*trianglecount].nx1 = read_signed_char(data,18);
		(*triangles)[*trianglecount].ny1 = read_signed_char(data,19);
		(*triangles)[*trianglecount].nz1 = read_signed_char(data,20);
		(*triangles)[*trianglecount].nx2 = read_signed_char(data,21);
		(*triangles)[*trianglecount].ny2 = read_signed_char(data,22);
		(*triangles)[*trianglecount].nz2 = read_signed_char(data,23);
		(*triangles)[*trianglecount].nx3 = read_signed_char(data,24);
		(*triangles)[*trianglecount].ny3 = read_signed_char(data,25);
		(*triangles)[*trianglecount].nz3 = read_signed_char(data,26);
		(*triangles)[*trianglecount].tx1 = read_signed_short(data,27);
		(*triangles)[*trianglecount].ty1 = read_signed_short(data,29);
		(*triangles)[*trianglecount].tx2 = read_signed_short(data,31);
		(*triangles)[*trianglecount].ty2 = read_signed_short(data,33);
		(*triangles)[*trianglecount].tx3 = read_signed_short(data,35);
		(*triangles)[*trianglecount].ty3 = read_signed_short(data,37);
		(*triangles)[*trianglecount].col1.r = read_unsigned_char(data,39);
		(*triangles)[*trianglecount].col1.g = read_unsigned_char(data,40);
		(*triangles)[*trianglecount].col1.b = read_unsigned_char(data,41);
		(*triangles)[*trianglecount].col1.a = read_unsigned_char(data,42);
		(*triangles)[*trianglecount].col2.r = read_unsigned_char(data,43);
		(*triangles)[*trianglecount].col2.g = read_unsigned_char(data,44);
		(*triangles)[*trianglecount].col2.b = read_unsigned_char(data,45);
		(*triangles)[*trianglecount].col2.a = read_unsigned_char(data,46);
		(*triangles)[*trianglecount].col3.r = read_unsigned_char(data,47);
		(*triangles)[*trianglecount].col3.g = read_unsigned_char(data,48);
		(*triangles)[*trianglecount].col3.b = read_unsigned_char(data,49);
		(*triangles)[*trianglecount].col3.a = read_unsigned_char(data,50);
		++*trianglecount;
	}
	
	fclose(f);
	return true;
	
}

// ########## BEGIN dll functions ##########

export double gmrender_texture_load(char* filename, char* filename_alpha) {
	unsigned short x,y;
	gmrender_texture* texture;
	imagestruct image,image_alpha;
	if(!image_load_bmp(&image,filename)) {
		return 0;
	}
	if(!image_load_bmp(&image_alpha,filename_alpha)) {
		image_free(&image);
		return 0;
	}
	if(image.width!=image_alpha.width || image.height!=image_alpha.height) {
		image_free(&image);
		image_free(&image_alpha);
		return 0;
	}
	for(y=0;y<image.height;++y) {
		for(x=0;x<image.width;++x) {
			image.data[y][x].a = (image_alpha.data[y][x].r+image_alpha.data[y][x].g+image_alpha.data[y][x].b)/3;
		}
	}
	texture = create_texture();
	if(texture==NULL) return 0;
	texture->id = ++id_textures;
	texture->image = image;
	image_free(&image_alpha);
	return texture->id;
}

export double gmrender_texture_destroy(double texture_id) {
	gmrender_texture* texture;
	if((texture=search_texture((unsigned long) texture_id))==NULL) return -1;
	image_free(&(texture->image));
	destroy_texture(texture);
	return 1;
}

export double gmrender_model_load(char* filename) {
	gmrender_model* model;
	gmrender_triangle* triangles;
	unsigned short trianglecount;
	if(!gmbin_load(&triangles,&trianglecount,filename)) return 0;
	model = create_model();
	if(model==NULL) return 0;
	model->id = ++id_models;
	model->triangles = triangles;
	model->trianglecount = trianglecount;
	model->texture = NULL;
	return model->id;
}

export double gmrender_model_set_material(double model_id, double texture_id, double tex_repeat, double tex_interpolate, double highlight_r, double highlight_g, double highlight_b, double specular_exponent) {
	gmrender_model* model;
	gmrender_texture* texture;
	if((model=search_model((unsigned long) model_id))==NULL) return -1;
	if((texture=search_texture((unsigned long) texture_id))==NULL) {
		model->texture = NULL;
	} else {
		model->texture = texture;
	}
	model->tex_repeat = (tex_repeat>=0.5);
	model->tex_interpolate = (tex_interpolate>=0.5);
	model->highlight_r = (unsigned char) highlight_r;
	model->highlight_g = (unsigned char) highlight_g;
	model->highlight_b = (unsigned char) highlight_b;
	model->specular_exponent = (unsigned short) specular_exponent;
	return 1;
}

export double gmrender_model_destroy(double model_id) {
	gmrender_model* model;
	if((model=search_model((unsigned long) model_id))==NULL) return -1;
	free(model->triangles);
	destroy_model(model);
	return 1;
}

export double gmrender_set_projection(double longitude, double latitude, double longitude_incr, double scale, double orig_x, double orig_y, double width, double height) {
	cam_longitude = (float) longitude;
	cam_latitude = (float) latitude;
	cam_longitude_incr = (float) longitude_incr;
	cam_scale = (float) scale;
	cam_orig_x = (unsigned short) orig_x;
	cam_orig_y = (unsigned short) orig_y;
	cam_width = (unsigned short) width;
	cam_height = (unsigned short) height;
	return 1;
}

export double gmrender_set_light(double ambient_r, double ambient_g, double ambient_b, double directional_r, double directional_g, double directional_b, double directional_longitude, double directional_latitude, double directional_longitude_incr) {
	light_ambient_r = (unsigned char) ambient_r;
	light_ambient_g = (unsigned char) ambient_g;
	light_ambient_b = (unsigned char) ambient_b;
	light_directional_r = (unsigned char) directional_r;
	light_directional_g = (unsigned char) directional_g;
	light_directional_b = (unsigned char) directional_b;
	light_directional_longitude = (float) directional_longitude;
	light_directional_latitude = (float) directional_latitude;
	light_directional_longitude_incr = (float) directional_longitude_incr;
	return 1;
}

export double gmrender_set_antialiasing(double samples, double jitter, double seed, double compalpha, double setedgepixels, double bg_r, double bg_g, double bg_b) {
	aa_samples = (unsigned short) samples;
	aa_jitter = (float) jitter;
	aa_seed = (unsigned long) seed;
	aa_compalpha = (compalpha>=0.5);
	aa_setedgepixels = (setedgepixels>=0.5);
	aa_bg_r = (unsigned char) bg_r;
	aa_bg_g = (unsigned char) bg_g;
	aa_bg_b = (unsigned char) bg_b;
}

#define edgepixel(xxx,yyy) {\
	if(image.data[yyy][xxx].a!=0) {\
		image.data[y][x].r = image.data[yyy][xxx].r;\
		image.data[y][x].g = image.data[yyy][xxx].g;\
		image.data[y][x].b = image.data[yyy][xxx].b;\
	}\
}

export double gmrender_render_strip(char* filename, char* filenamealpha, double frames, double frames_per_row) {
	unsigned short frame,frame_count,frame_mod,frame_x,frame_y,x,y,aax,aay,i,j,trianglecount,selectioncount,txx,tyy;
	float cos_yaw,sin_yaw,cos_pitch,sin_pitch,xfrom,yfrom,zfrom,xdown,ydown,zdown,xright,yright,zright,light_x,light_y,light_z,nx,ny,nz,d,xx,yy,pixel_r,pixel_g,pixel_b,pixel_a,total_r,total_g,total_b,total_a,p,q,z,r,g,b,a,tx,ty,light1,light2,best_z,max_z;
	gmrender_model* model;
	gmrender_triangle triangle;
	gmrender_triangle2 tr;
	gmrender_triangle2 *triangles,*triangleselection;
	imagestruct image;
	
	frame_count = (unsigned short) frames;
	frame_mod = (unsigned short) frames_per_row;
	
	if(!image_load_empty(&image,
	cam_width*((frame_count<frame_mod)?(frame_count):(frame_mod)),
	cam_height*(frame_count+((frame_count%frame_mod!=0)?(frame_mod-(frame_count%frame_mod)):(0)))/frame_mod,0,0,0,0)) {
		return 0;
	}
	
	trianglecount = 0;
	for(model=models;model!=NULL;model=model->next) {
		trianglecount += model->trianglecount;
	}
	triangles = (gmrender_triangle2*) malloc(trianglecount*sizeof(gmrender_triangle2));
	if(triangles==NULL) {
		image_free(&image);
		return 0;
	}
	triangleselection = (gmrender_triangle2*) malloc(trianglecount*sizeof(gmrender_triangle2));
	if(triangleselection==NULL) {
		free(triangles);
		image_free(&image);
		return 0;
	}
	
	srand(aa_seed);
	
	for(frame=0;frame<frame_count;++frame) {
		
		cos_yaw = cos((cam_longitude+float(frame)*cam_longitude_incr)*M_PI/180.0);
		sin_yaw = sin((cam_longitude+float(frame)*cam_longitude_incr)*M_PI/180.0);
		cos_pitch = cos((cam_latitude)*M_PI/180.0);
		sin_pitch = sin((cam_latitude)*M_PI/180.0);
		xfrom = -cos_yaw*cos_pitch;
		yfrom = sin_yaw*cos_pitch;
		zfrom = -sin_pitch;
		xdown = cos_yaw*sin_pitch;
		ydown = -sin_yaw*sin_pitch;
		zdown = -cos_pitch;
		xright = sin_yaw;
		yright = cos_yaw;
		zright = 0.0;
		
		cos_yaw = cos((light_directional_longitude+float(frame)*light_directional_longitude_incr)*M_PI/180.0);
		sin_yaw = sin((light_directional_longitude+float(frame)*light_directional_longitude_incr)*M_PI/180.0);
		cos_pitch = cos((light_directional_latitude)*M_PI/180.0);
		sin_pitch = sin((light_directional_latitude)*M_PI/180.0);
		light_x = cos_yaw*cos_pitch;
		light_y = -sin_yaw*cos_pitch;
		light_z = sin_pitch;
		
		i = 0;
		for(model=models;model!=NULL;model=model->next) {
			for(j=0;j<model->trianglecount;++j) {
				triangle = model->triangles[j];
				triangles[i].x1 = (xright*float(triangle.x1)/10+yright*float(triangle.y1)/10+zright*float(triangle.z1)/10)*cam_scale;
				triangles[i].y1 = (xdown *float(triangle.x1)/10+ydown *float(triangle.y1)/10+zdown *float(triangle.z1)/10)*cam_scale;
				triangles[i].z1 = (xfrom *float(triangle.x1)/10+yfrom *float(triangle.y1)/10+zfrom *float(triangle.z1)/10)*cam_scale;
				triangles[i].x2 = (xright*float(triangle.x2)/10+yright*float(triangle.y2)/10+zright*float(triangle.z2)/10)*cam_scale;
				triangles[i].y2 = (xdown *float(triangle.x2)/10+ydown *float(triangle.y2)/10+zdown *float(triangle.z2)/10)*cam_scale;
				triangles[i].z2 = (xfrom *float(triangle.x2)/10+yfrom *float(triangle.y2)/10+zfrom *float(triangle.z2)/10)*cam_scale;
				triangles[i].x3 = (xright*float(triangle.x3)/10+yright*float(triangle.y3)/10+zright*float(triangle.z3)/10)*cam_scale;
				triangles[i].y3 = (xdown *float(triangle.x3)/10+ydown *float(triangle.y3)/10+zdown *float(triangle.z3)/10)*cam_scale;
				triangles[i].z3 = (xfrom *float(triangle.x3)/10+yfrom *float(triangle.y3)/10+zfrom *float(triangle.z3)/10)*cam_scale;
				triangles[i].nx1 = float(triangle.nx1)/100;
				triangles[i].ny1 = float(triangle.ny1)/100;
				triangles[i].nz1 = float(triangle.nz1)/100;
				triangles[i].nx2 = float(triangle.nx2)/100;
				triangles[i].ny2 = float(triangle.ny2)/100;
				triangles[i].nz2 = float(triangle.nz2)/100;
				triangles[i].nx3 = float(triangle.nx3)/100;
				triangles[i].ny3 = float(triangle.ny3)/100;
				triangles[i].nz3 = float(triangle.nz3)/100;
				triangles[i].tx1 = float(triangle.tx1)/1000;
				triangles[i].ty1 = float(triangle.ty1)/1000;
				triangles[i].tx2 = float(triangle.tx2)/1000;
				triangles[i].ty2 = float(triangle.ty2)/1000;
				triangles[i].tx3 = float(triangle.tx3)/1000;
				triangles[i].ty3 = float(triangle.ty3)/1000;
				triangles[i].r1 = float(triangle.col1.r)/255;
				triangles[i].g1 = float(triangle.col1.g)/255;
				triangles[i].b1 = float(triangle.col1.b)/255;
				triangles[i].a1 = float(triangle.col1.a)/255;
				triangles[i].add_r1 = 4.0*float(model->highlight_r)/255;
				triangles[i].add_g1 = 4.0*float(model->highlight_g)/255;
				triangles[i].add_b1 = 4.0*float(model->highlight_b)/255;
				triangles[i].r2 = float(triangle.col2.r)/255;
				triangles[i].g2 = float(triangle.col2.g)/255;
				triangles[i].b2 = float(triangle.col2.b)/255;
				triangles[i].a2 = float(triangle.col2.a)/255;
				triangles[i].add_r2 = 4.0*float(model->highlight_r)/255;
				triangles[i].add_g2 = 4.0*float(model->highlight_g)/255;
				triangles[i].add_b2 = 4.0*float(model->highlight_b)/255;
				triangles[i].r3 = float(triangle.col3.r)/255;
				triangles[i].g3 = float(triangle.col3.g)/255;
				triangles[i].b3 = float(triangle.col3.b)/255;
				triangles[i].a3 = float(triangle.col3.a)/255;
				triangles[i].add_r3 = 4.0*float(model->highlight_r)/255;
				triangles[i].add_g3 = 4.0*float(model->highlight_g)/255;
				triangles[i].add_b3 = 4.0*float(model->highlight_b)/255;
				triangles[i].specular_exponent = model->specular_exponent;
				triangles[i].texture = model->texture;
				triangles[i].tex_repeat = model->tex_repeat;
				triangles[i].tex_interpolate = model->tex_interpolate;
				++i;
			}
		}
		
		frame_x = cam_width*(frame%frame_mod);
		frame_y = cam_height*(frame-(frame%frame_mod))/frame_mod;
		for(y=0;y<cam_height;++y) {
			
			yy = float(y-cam_orig_y);
			
			selectioncount = 0;
			for(i=0;i<trianglecount;++i) {
				tr = triangles[i];
				if(min(min(tr.y1,tr.y2),tr.y3)<=yy+1.0 && max(max(tr.y1,tr.y2),tr.y3)>=yy && (tr.y1-tr.y2)*(tr.x3-tr.x1)+(tr.x2-tr.x1)*(tr.y3-tr.y1)>0.001) {
					triangleselection[selectioncount] = tr;
					++selectioncount;
				}
			}
			
			for(x=0;x<cam_width;++x) {
				
				total_r = 0.0;
				total_g = 0.0;
				total_b = 0.0;
				total_a = 0.0;
				
				for(aay=0;aay<aa_samples;++aay) {
					for(aax=0;aax<aa_samples;++aax) {
						
						xx = float(x-cam_orig_x)+(float(aax)+0.5-aa_jitter/2+aa_jitter*rand()/RAND_MAX)/float(aa_samples);
						yy = float(y-cam_orig_y)+(float(aay)+0.5-aa_jitter/2+aa_jitter*rand()/RAND_MAX)/float(aa_samples);
						max_z = 10000.0;
						pixel_r = 0.0;
						pixel_g = 0.0;
						pixel_b = 0.0;
						pixel_a = 0.0;
						
						while(true) {
							
							best_z = -10000.0;
							
							for(i=0;i<selectioncount;++i) {
								tr = triangleselection[i];
								if((tr.y1-tr.y2)*(xx-tr.x1)+(tr.x2-tr.x1)*(yy-tr.y1)>=0
								&& (tr.y2-tr.y3)*(xx-tr.x2)+(tr.x3-tr.x2)*(yy-tr.y2)>=0
								&& (tr.y3-tr.y1)*(xx-tr.x3)+(tr.x1-tr.x3)*(yy-tr.y3)>=0 ) {
									
									p = ((xx-tr.x1)*(tr.y3-tr.y1)-(yy-tr.y1)*(tr.x3-tr.x1))/((tr.x2-tr.x1)*(tr.y3-tr.y1)-(tr.y2-tr.y1)*(tr.x3-tr.x1));
									q = ((xx-tr.x1)*(tr.y2-tr.y1)-(yy-tr.y1)*(tr.x2-tr.x1))/((tr.x3-tr.x1)*(tr.y2-tr.y1)-(tr.y3-tr.y1)*(tr.x2-tr.x1));
									
									z = tr.z1+p*(tr.z2-tr.z1)+q*(tr.z3-tr.z1);
									if(z>best_z && z<max_z) {
										best_z = z;
										
										r = tr.r1+p*(tr.r2-tr.r1)+q*(tr.r3-tr.r1);
										g = tr.g1+p*(tr.g2-tr.g1)+q*(tr.g3-tr.g1);
										b = tr.b1+p*(tr.b2-tr.b1)+q*(tr.b3-tr.b1);
										a = tr.a1+p*(tr.a2-tr.a1)+q*(tr.a3-tr.a1);
										
										if(tr.texture!=NULL) {
											tx = (tr.tx1+p*(tr.tx2-tr.tx1)+q*(tr.tx3-tr.tx1))*float(tr.texture->image.width)-0.5;
											ty = (tr.ty1+p*(tr.ty2-tr.ty1)+q*(tr.ty3-tr.ty1))*float(tr.texture->image.height)-0.5;
											if(!tr.tex_repeat) {
												tx = min(float(tr.texture->image.width-1),max(0.0,tx));
												ty = min(float(tr.texture->image.height-1),max(0.0,ty));
											}
											if(tr.tex_interpolate) {
												txx = (unsigned short) floor(tx);
												tyy = (unsigned short) floor(ty);
												tx -= floor(tx);
												ty -= floor(ty);
											} else {
												txx = (unsigned short) round(tx);
												tyy = (unsigned short) round(ty);
												tx = 0.0;
												ty = 0.0;
											}
											r *= 
											(1.0-ty)*(1.0-tx)*float(tr.texture->image.data[(tyy)%tr.texture->image.height][(txx)%tr.texture->image.width].r)/255.0+
											(1.0-ty)*(tx)*float(tr.texture->image.data[(tyy)%tr.texture->image.height][(txx+1)%tr.texture->image.width].r)/255.0+
											(ty)*(1.0-tx)*float(tr.texture->image.data[(tyy+1)%tr.texture->image.height][(txx)%tr.texture->image.width].r)/255.0+
											(ty)*(tx)*float(tr.texture->image.data[(tyy+1)%tr.texture->image.height][(txx+1)%tr.texture->image.width].r)/255.0;
											g *= 
											(1.0-ty)*(1.0-tx)*float(tr.texture->image.data[(tyy)%tr.texture->image.height][(txx)%tr.texture->image.width].g)/255.0+
											(1.0-ty)*(tx)*float(tr.texture->image.data[(tyy)%tr.texture->image.height][(txx+1)%tr.texture->image.width].g)/255.0+
											(ty)*(1.0-tx)*float(tr.texture->image.data[(tyy+1)%tr.texture->image.height][(txx)%tr.texture->image.width].g)/255.0+
											(ty)*(tx)*float(tr.texture->image.data[(tyy+1)%tr.texture->image.height][(txx+1)%tr.texture->image.width].g)/255.0;
											b *= 
											(1.0-ty)*(1.0-tx)*float(tr.texture->image.data[(tyy)%tr.texture->image.height][(txx)%tr.texture->image.width].b)/255.0+
											(1.0-ty)*(tx)*float(tr.texture->image.data[(tyy)%tr.texture->image.height][(txx+1)%tr.texture->image.width].b)/255.0+
											(ty)*(1.0-tx)*float(tr.texture->image.data[(tyy+1)%tr.texture->image.height][(txx)%tr.texture->image.width].b)/255.0+
											(ty)*(tx)*float(tr.texture->image.data[(tyy+1)%tr.texture->image.height][(txx+1)%tr.texture->image.width].b)/255.0;
											a *= 
											(1.0-ty)*(1.0-tx)*float(tr.texture->image.data[(tyy)%tr.texture->image.height][(txx)%tr.texture->image.width].a)/255.0+
											(1.0-ty)*(tx)*float(tr.texture->image.data[(tyy)%tr.texture->image.height][(txx+1)%tr.texture->image.width].a)/255.0+
											(ty)*(1.0-tx)*float(tr.texture->image.data[(tyy+1)%tr.texture->image.height][(txx)%tr.texture->image.width].a)/255.0+
											(ty)*(tx)*float(tr.texture->image.data[(tyy+1)%tr.texture->image.height][(txx+1)%tr.texture->image.width].a)/255.0;
										}
										
										nx = tr.nx1+p*(tr.nx2-tr.nx1)+q*(tr.nx3-tr.nx1);
										ny = tr.ny1+p*(tr.ny2-tr.ny1)+q*(tr.ny3-tr.ny1);
										nz = tr.nz1+p*(tr.nz2-tr.nz1)+q*(tr.nz3-tr.nz1);
										
										d = sqrt(nx*nx+ny*ny+nz*nz);
										if(total_a>0.001) {
											nx /= d;
											ny /= d;
											nz /= d;
										}
										
										light1 = -light_x*nx-light_y*ny-light_z*nz;
										light2 = (light_x+2.0*light1*nx)*xfrom+(light_y+2.0*light1*ny)*yfrom+(light_z+2.0*light1*nz)*zfrom;
										light1 = max(0.0,light1);
										light2 = pow(max(0.0,light2),tr.specular_exponent);
										
										r *= float(light_ambient_r)/255+float(light_directional_r)/255*light1;
										g *= float(light_ambient_r)/255+float(light_directional_r)/255*light1;
										b *= float(light_ambient_r)/255+float(light_directional_r)/255*light1;
										
										r += (tr.add_r1+p*(tr.add_r2-tr.add_r1)+q*(tr.add_r3-tr.add_r1))*float(light_directional_r)/255*light2;
										g += (tr.add_g1+p*(tr.add_g2-tr.add_g1)+q*(tr.add_g3-tr.add_g1))*float(light_directional_g)/255*light2;
										b += (tr.add_b1+p*(tr.add_b2-tr.add_b1)+q*(tr.add_b3-tr.add_b1))*float(light_directional_b)/255*light2;
										
									}
									
								}
							}
							
							if(best_z==-10000.0) break;
							max_z = best_z-0.01;
							
							pixel_r += r*a*(1.0-pixel_a);
							pixel_g += g*a*(1.0-pixel_a);
							pixel_b += b*a*(1.0-pixel_a);
							pixel_a += a*(1.0-pixel_a);
							
							if(pixel_a>0.999) break;
							
						}
						
						total_r += pixel_r;
						total_g += pixel_g;
						total_b += pixel_b;
						total_a += pixel_a;
						
					}
				}
				
				total_r /= float(aa_samples)*float(aa_samples);
				total_g /= float(aa_samples)*float(aa_samples);
				total_b /= float(aa_samples)*float(aa_samples);
				total_a /= float(aa_samples)*float(aa_samples);
				
				if(total_a>0.001) {
					if(aa_compalpha) {
						total_r /= total_a;
						total_g /= total_a;
						total_b /= total_a;
					}
					image.data[frame_y+y][frame_x+x].r = (unsigned char) (max(0,min(255,round(255.0*total_r))));
					image.data[frame_y+y][frame_x+x].g = (unsigned char) (max(0,min(255,round(255.0*total_g))));
					image.data[frame_y+y][frame_x+x].b = (unsigned char) (max(0,min(255,round(255.0*total_b))));
					image.data[frame_y+y][frame_x+x].a = (unsigned char) (max(0,min(255,round(255.0*total_a))));
				} else {
					image.data[frame_y+y][frame_x+x].r = aa_bg_r;
					image.data[frame_y+y][frame_x+x].g = aa_bg_g;
					image.data[frame_y+y][frame_x+x].b = aa_bg_b;
					image.data[frame_y+y][frame_x+x].a = 0;
				}
				
			}
		}
		
	}
	
	if(aa_setedgepixels) {
		for(y=0;y<image.height;++y) {
			for(x=0;x<image.width;++x) {
				if(image.data[y][x].a==0) {
					if(1<=x && 1<=y) edgepixel(x-1,y-1);
					if(x<image.width-1 && 1<=y) edgepixel(x+1,y-1);
					if(1<=x && y<image.height-1) edgepixel(x-1,y+1);
					if(x<image.width-1 && y<image.height-1) edgepixel(x+1,y+1);
					if(1<=x) edgepixel(x-1,y);
					if(x<image.width-1) edgepixel(x+1,y);
					if(1<=y) edgepixel(x,y-1);
					if(y<image.height-1) edgepixel(x,y+1);
				}
			}
		}
	}
	if(!image_save_bmp(&image,filename,aa_compalpha)) {
		free(triangles);
		free(triangleselection);
		image_free(&image);
		return 0;
	}
	if(filenamealpha[0]!='\0') {
		for(y=0;y<image.height;++y) {
			for(x=0;x<image.width;++x) {
				image.data[y][x].r = image.data[y][x].a;
				image.data[y][x].g = image.data[y][x].a;
				image.data[y][x].b = image.data[y][x].a;
			}
		}
		if(!image_save_bmp(&image,filenamealpha,false)) {
			free(triangles);
			free(triangleselection);
			image_free(&image);
			return 0;
		}
	}
	
	free(triangles);
	free(triangleselection);
	image_free(&image);
	
	return 1;
}

